package org.eclipse.swt.custom;

import java.util.*;

import org.eclipse.swt.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.widgets.*;

/**
 * @author Marc Strapetz
 */
public class QStyledText extends StyledText {

	// Fields =================================================================

	private final Map<Integer, Action> keyToAction = new HashMap<>();

	private int[] lineSpacings;
	private int[] cachedLineYOffsets;

	// Setup ==================================================================

	public QStyledText(Composite parent, int style) {
		super(parent, style);

		renderer.dispose();
		renderer = new Renderer(getDisplay(), this);
		renderer.setContent(content);
		renderer.setFont(getFont(), tabLength);

		final Listener listener = event -> {
			if (event.type == SWT.Verify) {
				invalidateCaches();
				return;
			}

			if (event.type == ST.VerifyKey) {
				if (!event.doit) {
					return;
				}

				if ((event.keyCode & SWT.KEY_MASK) == 0) {
					return;
				}

				final Action action = getAction(event);
				if (action != null) {
					if (action.perform(this, event)) {
						event.doit = false;
					}
				}
			}
		};
		addListener(SWT.Verify, listener);
		addListener(ST.VerifyKey, listener);
	}

	// Implemented ============================================================

	@Override
	public void setText(String text) {
		super.setText(text);
		invalidateCaches();
	}

	// Accessing ==============================================================

	public int getLineHeightAndSpacing(int lineIndex) {
		return renderer.getLineHeight(lineIndex);
	}

	public int getLineSpacing(int lineIndex) {
		if (lineSpacings == null || lineIndex >= lineSpacings.length) {
			return getLineSpacing();
		}

		return lineSpacings[lineIndex];
	}

	public void setLineSpacing(int lineIndex, int lineSpacing) {
		setVariableLineHeight();

		final int lineCount = getLineCount();
		if (lineSpacings == null) {
			lineSpacings = new int[lineCount];
		}
		else if (lineSpacings.length != lineCount) {
			final int[] newLineSpacings = new int[lineCount];
			System.arraycopy(lineSpacings, 0, newLineSpacings, 0, Math.min(lineSpacings.length, newLineSpacings.length));
			lineSpacings = newLineSpacings;
		}

		lineSpacings[lineIndex] = lineSpacing;
	}

	public void clearLineSpacings() {
		lineSpacings = null;

		if (fixedLineHeight) {
			return;
		}
		fixedLineHeight = true;
		renderer.calculateIdle();
	}

	public final int getLineHeight(int lineIndex, boolean withSpacing) {
		if (!refreshCaches()) {
			return getLineHeight() + (withSpacing ? getLineSpacing() : 0);
		}

		int height = cachedLineYOffsets[lineIndex + 1] - cachedLineYOffsets[lineIndex];
		if (!withSpacing) {
			height -= lineSpacings != null && lineIndex < lineSpacings.length ? lineSpacings[lineIndex] : getLineSpacing();
		}

		return height;
	}

	public int getOverallHeight() {
		final int lineCount = getLineCount();
		if (!refreshCaches()) {
			return (getLineHeight() + getLineSpacing()) * lineCount;
		}

		return cachedLineYOffsets[lineCount];
	}

	/**
	 * @see #org.eclipse.swt.custom.StyledText#setKeyBinding(int, int)
	 */
	public void registerAction(int key, Action action) {
		if (action == null) {
			throw new NullPointerException();
		}

		final int modifiers = key & SWT.MODIFIER_MASK;
		final int keyInt = key & SWT.KEY_MASK;
		final char keyChr = (char)keyInt;

		if (Character.isDefined(keyInt) && Character.isLetter(keyChr)) {
			// make the keybinding case insensitive by adding it
			// in its upper and lower case form
			char chr = Character.toUpperCase(keyChr);
			int newKey = chr | modifiers;
			keyToAction.put(newKey, action);

			chr = Character.toLowerCase(keyChr);
			newKey = chr | modifiers;
			keyToAction.put(newKey, action);
		}
		else {
			keyToAction.put(key, action);
		}

		setKeyBinding(key, SWT.NULL);
	}

	public Action getAction(Event event) {
		return keyToAction.get(event.keyCode | event.stateMask);
	}

	public String getLineDelimiter(int lineIndex) {
		final StyledTextContent content = getContent();
		final int lineCount = content.getLineCount();
		if (lineIndex < 0 || lineIndex >= lineCount - 1) {
			return "";
		}

		// this should be more efficient than creating strings for the whole line to determine the line length
		final int currentLineStart = content.getOffsetAtLine(lineIndex);
		final int nextLineStart = content.getOffsetAtLine(lineIndex + 1);
		final int start = Math.max(currentLineStart, nextLineStart - 2);
		final String lastCharacters = content.getTextRange(start, nextLineStart - start);
		for (int i = lastCharacters.length(); i-- > 0; ) {
			final char chr = lastCharacters.charAt(i);
			if (chr != '\n' && chr != '\r') {
				return lastCharacters.substring(i + 1);
			}
		}
		return lastCharacters;
	}

	// Utils ==================================================================

	private boolean refreshCaches() {
		if (lineSpacings == null) {
			cachedLineYOffsets = null;
			return false;
		}

		if (cachedLineYOffsets != null) {
			return true;
		}

		final int lineCount = getLineCount();
		cachedLineYOffsets = new int[lineCount + 1];

		cachedLineYOffsets[0] = 0;
		for (int lineIndex = 0; lineIndex < lineCount; lineIndex++) {
			final int lineHeight = getLineHeight(getOffsetAtLine(lineIndex));
			final int lineSpacing = lineIndex < lineSpacings.length ? lineSpacings[lineIndex] : getLineSpacing();
			cachedLineYOffsets[lineIndex + 1] = cachedLineYOffsets[lineIndex] + lineHeight + lineSpacing;
		}

		return true;
	}

	private void invalidateCaches() {
		cachedLineYOffsets = null;
	}

	// Inner Classes ==========================================================

	private class Renderer extends StyledTextRenderer {
		private Renderer(Device device, StyledText styledText) {
			super(device, styledText);
		}

		@Override
		TextLayout getTextLayout(int lineIndex) {
			int lineSpacing = styledText.lineSpacing;
			if (lineSpacings != null && lineIndex < lineSpacings.length) {
				lineSpacing = lineSpacings[lineIndex];
			}

			return getTextLayout(lineIndex, getOrientation(), getWrapWidth(), lineSpacing);
		}
	}

	public interface Action {
		boolean perform(StyledText text, Event event);
	}
}
